program eddy_graph

  use, intrinsic:: ISO_FORTRAN_ENV

  ! Libraries:
  use ezmpi, only: ezmpi_bcast
  use jumble, only: get_command_arg_dyn, new_unit, ediff1d, assert, &
       read_opcol
  use mpi_f08, only: mpi_init, mpi_finalize, MPI_Comm_rank, MPI_Comm_world, &
       MPI_Comm_size
  use shapelib_03, only: dbf_read_attribute_03

  use config_graph_m, only: config_graph, cyclone, max_delta
  use derived_types, only: snapshot, shpc_slice_handler, shpc_slice_meta
  use dispatch_snapshot_m, only: dispatch_snapshot
  use get_snapshot_m, only: get_snapshot
  use overlap_m, only: overlap
  use read_grid_m, only: read_grid
  use unit_edge_m, only: open_edge_file, unit_edge
  use shpc_open_m, only: shpc_open
  use shpc_close_m, only: shpc_close

  implicit none

  integer rank, n_proc, delta, i, j, iostat
  integer:: n_dates = huge(0) ! number of dates to read
  integer k ! date index
  integer k_begin, k_end, k_end_main_loop
  character(len = :), allocatable:: shpc_dir
  integer unit_isolated, unit
  integer n_slices ! number of slices in the input SHPC
  integer e_overestim ! over-estimation of the number of eddies at each date
  TYPE(shpc_slice_handler), allocatable:: hshpc(:) ! (n_slices)
  type(shpc_slice_meta), allocatable:: ssm(:) ! (n_slices)
  type(snapshot), allocatable:: flow(:) ! (max_delta + 1)
  character(len = 30) file

  namelist /main_nml/ n_dates

  !-------------------------------------------------------------------------

  ! 1. Front matter

  call mpi_init
  call MPI_Comm_rank(MPI_Comm_world, rank)
  call MPI_Comm_size(MPI_Comm_world, n_proc)
  call get_command_arg_dyn(1, shpc_dir, "Required argument: SHPC-directory")
  call read_grid(rank, shpc_dir)
  call config_graph(rank)

  if (rank == 0) then
     ! n_slices:

     call new_unit(unit)
     open(unit, file = shpc_dir // "/n_slices.txt", status = "old", &
          action = "read", position = "rewind", iostat = iostat)

     if (iostat == 0) then
        read(unit, fmt = *) n_slices
        close(unit)
     else
        print *, "eddy_graph: Could not open ", shpc_dir, "/n_slices.txt"
        print *, "eddy_graph: Assuming there is only one slice..."
        n_slices = 1
     end if

     ! main_nml:
     write(unit = error_unit, nml = main_nml)
     write(unit = error_unit, fmt = *) "Enter namelist main_nml."
     read(unit = *, nml = main_nml)
     write(unit = *, nml = main_nml)
     call assert(n_dates == huge(0) .or. n_slices == 1, "If we have several " &
          // "slices then we read all the dates from them")
  end if

  call ezmpi_bcast(n_slices, root = 0)
  allocate(hshpc(n_slices), ssm(n_slices))

  do i = 1, n_slices
     call shpc_open(hshpc(i), trim(shpc_dir), cyclone = cyclone, &
          slice = i - 1, pszaccess = "rb")
     call dbf_read_attribute_03(ssm(i)%d0, hshpc(i)%extremum, &
          hshpc(i)%extr_date, ishape = 0)
     call read_opcol(ssm(i)%ishape_last, hshpc(i)%unit, my_lbound = ssm(i)%d0)
  end do

  if (rank == 0) then
     if (n_dates == huge(0)) &
          n_dates = sum([(size(ssm(i)%ishape_last), i = 1, n_slices)])
     print *, "n_dates = ", n_dates
     call assert(n_dates >= max_delta + 1, &
          "eddy_graph: n_dates should be >= max_delta + 1")
     call assert(n_proc <= n_dates / (max_delta + 1), &
          "eddy_graph: n_proc should be <= n_dates / (max_delta + 1)")
     e_overestim = maxval([(ssm(i)%ishape_last(ssm(i)%d0) + 1, &
          ediff1d(ssm(i)%ishape_last), i = 1, n_slices)])
     open(unit, file = "e_overestim.txt", status = "replace", action = "write")
     write(unit, fmt = *) e_overestim
     close(unit)
  end if

  call ezmpi_bcast(n_dates, root = 0)
  call ezmpi_bcast(e_overestim, root = 0)

  ! Open output files:

  call new_unit(unit_isolated)
  write(unit = file, fmt = '("isolated_nodes_", i0, ".txt")') rank
  open(unit_isolated, file = trim(file), status = "replace", action = "write")

  call open_edge_file(rank)
  ! Note: no title line, because we want the edgelist to be easily
  ! readable by various graph software.

  k_begin = ssm(1)%d0 + (rank * n_dates) / n_proc

  if (rank < n_proc - 1) then
     k_end = ssm(1)%d0 + ((rank + 1) * n_dates) / n_proc + max_delta - 1
     k_end_main_loop = k_end - max_delta + 1
  else
     k_end = ssm(1)%d0 + n_dates - 1
     k_end_main_loop = k_end
  end if

  allocate(flow(max_delta + 1))

  ! 2. Prologue:

  do k = k_begin, k_begin + max_delta
     call get_snapshot(flow(k - k_begin + 1), hshpc, ssm, k, k_end, rank, &
          n_proc) ! (read)
  end do

  do delta = 1, max_delta
     do k = k_begin + delta, k_begin + max_delta
        call overlap(flow, e_overestim, k, delta, j = k - k_begin + 1)
     end do
  end do

  ! 3. Main loop

  do k = k_begin + max_delta + 1, k_end_main_loop
     call dispatch_snapshot(flow(1), unit_isolated, rank, k_begin, &
          k = k - max_delta - 1)
     flow(:max_delta) = flow(2:)
     call get_snapshot(flow(max_delta + 1), hshpc, ssm, k, k_end, rank, n_proc)

     do delta = 1, max_delta
        call overlap(flow, e_overestim, k, delta, j = max_delta + 1)
     end do
  end do

  ! 4. Epilogue

  do k = k_end_main_loop + 1, k_end
     ! {rank < n_proc - 1 and k >= k_end - max_delta + 2}
     call dispatch_snapshot(flow(1), unit_isolated, rank, k_begin, &
          k = k - max_delta - 1)
     flow(:max_delta) = flow(2:)
     call get_snapshot(flow(max_delta + 1), hshpc, ssm, k, k_end, rank, n_proc)
     ! (reception)

     ! Stitching:
     do delta = k - k_end + max_delta, max_delta
        call overlap(flow, e_overestim, k, delta, j = max_delta + 1)
     end do
  end do

  do j = 1, max_delta + 1
     call dispatch_snapshot(flow(j), unit_isolated, rank, k_begin, &
          k = k_end - max_delta - 1 + j)
  end do

  ! 5. Back matter

  do i = 1, n_slices
     CALL shpc_close(hshpc(i))
  end do

  close(unit_isolated)
  close(unit_edge)
  call mpi_finalize

end program eddy_graph
